/* ========================================================================= *
 *                                                                           *
 *                 The Apache Software License,  Version 1.1                 *
 *                                                                           *
 *             Copyright (c) 2002 The Apache Software Foundation.            *
 *                           All rights reserved.                            *
 *                                                                           *
 * ========================================================================= *
 *                                                                           *
 * Redistribution and use in source and binary forms,  with or without modi- *
 * fication, are permitted provided that the following conditions are met:   *
 *                                                                           *
 * 1. Redistributions of source code  must retain the above copyright notice *
 *    notice, this list of conditions and the following disclaimer.          *
 *                                                                           *
 * 2. Redistributions  in binary  form  must  reproduce the  above copyright *
 *    notice,  this list of conditions  and the following  disclaimer in the *
 *    documentation and/or other materials provided with the distribution.   *
 *                                                                           *
 * 3. The end-user documentation  included with the redistribution,  if any, *
 *    must include the following acknowlegement:                             *
 *                                                                           *
 *       "This product includes  software developed  by the Apache  Software *
 *        Foundation <http://www.apache.org/>."                              *
 *                                                                           *
 *    Alternately, this acknowlegement may appear in the software itself, if *
 *    and wherever such third-party acknowlegements normally appear.         *
 *                                                                           *
 * 4. The names  "The Jakarta  Project",  and  "Apache  Software Foundation" *
 *    must not  be used  to endorse  or promote  products derived  from this *
 *    software without  prior written  permission.  For written  permission, *
 *    please contact <apache@apache.org>.                                    *
 *                                                                           *
 * 5. Products derived from this software may not be called "Apache" nor may *
 *    "Apache" appear in their names without prior written permission of the *
 *    Apache Software Foundation.                                            *
 *                                                                           *
 * THIS SOFTWARE IS PROVIDED "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES *
 * INCLUDING, BUT NOT LIMITED TO,  THE IMPLIED WARRANTIES OF MERCHANTABILITY *
 * AND FITNESS FOR  A PARTICULAR PURPOSE  ARE DISCLAIMED.  IN NO EVENT SHALL *
 * THE APACHE  SOFTWARE  FOUNDATION OR  ITS CONTRIBUTORS  BE LIABLE  FOR ANY *
 * DIRECT,  INDIRECT,   INCIDENTAL,  SPECIAL,  EXEMPLARY,  OR  CONSEQUENTIAL *
 * DAMAGES (INCLUDING,  BUT NOT LIMITED TO,  PROCUREMENT OF SUBSTITUTE GOODS *
 * OR SERVICES;  LOSS OF USE,  DATA,  OR PROFITS;  OR BUSINESS INTERRUPTION) *
 * HOWEVER CAUSED AND  ON ANY  THEORY  OF  LIABILITY,  WHETHER IN  CONTRACT, *
 * STRICT LIABILITY, OR TORT  (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN *
 * ANY  WAY  OUT OF  THE  USE OF  THIS  SOFTWARE,  EVEN  IF  ADVISED  OF THE *
 * POSSIBILITY OF SUCH DAMAGE.                                               *
 *                                                                           *
 * ========================================================================= *
 *                                                                           *
 * This software  consists of voluntary  contributions made  by many indivi- *
 * duals on behalf of the  Apache Software Foundation.  For more information *
 * on the Apache Software Foundation, please see <http://www.apache.org/>.   *
 *                                                                           *
 * ========================================================================= */

package org.apache.commons.launcher;

import java.io.BufferedInputStream;
import java.io.InputStream;
import java.io.IOException;

 /**
 * A class that prevents blocking reads. This class is usually used to prevent
 * native blocking reads of System.in on Windows platforms since on some
 * Windows platforms, reading System.in will cause the entire JVM to block.
 *
 * @author Patrick Luby
 */
public class NonBlockingInputStream extends BufferedInputStream {

    //------------------------------------------------------------ Constructors

    /**
     * Creates a <code>NonBlockingInputStream</code> by using the
     * {@link BufferedInputStream} as the underlying input stream.
     *
     * @param in the underlying input stream
     */
    public NonBlockingInputStream(InputStream in) {

        super(in);

    }

    //----------------------------------------------------------------- Methods

    /**
     * Reads the next byte of data from this input stream. This method invokes
     * the superclass' {@link BufferedInputStream#read()} method if the number
     * of available bytes is greater than 0.
     *
     * @return the next byte of data, or -1 if the end of the stream is reached
     * @throws IOExeption if an I/O error occurs
     */
    public int read() throws IOException {

        byte[] b = new byte[1];
        int bytesRead = readWhenAvailable(b, 0, 1);
        if (bytesRead == -1)
            return bytesRead;
        else 
           return (int)b[0];

    }

    /**
     * Reads some number of bytes from the input stream and stores them into the     * specified buffer. This method invokes the superclass'
     * {@link BufferedInputStream#read(byte[])} method if the number of
     * available bytes is greater than 0.
     *
     * @param b the buffer into which the data is read
     * @return the total number of bytes read into the buffer, or -1 if the end
     *  of the stream is reached
     * @throws IOExeption if an I/O error occurs
     */
    public int read(byte[] b) throws IOException {

        return readWhenAvailable(b, 0, b.length);

    }

    /**
     * Reads up to <code>len</code> bytes from the input stream and stores them
     * into the specified buffer. This method invokes the superclass'
     * {@link BufferedInputStream#read(byte[], int, int)} method if the number
     * of available bytes is greater than 0.
     *
     * @param b the buffer into which the data is read
     * @param off the start offset in array b at which the data is written
     * @param len the maximum number of bytes to read 
     * @return the total number of bytes read into the buffer, or -1 if the end
     *  of the stream is reached
     * @throws IOExeption if an I/O error occurs
     */
    public int read(byte[] b, int off, int len)
        throws IOException
    {

        return readWhenAvailable(b, off, len);

    }

    /**
     * Skips over and discards <code>n</code> bytes of data from this input
     * stream. This method invokes the superclass'
     * {@link BufferedInputStream#skip(long)} method if the number of
     * available bytes is greater than 0.
     *
     * @param n the number of bytes to be skipped
     * @return the actual number of bytes skipped
     * @throws IOExeption if an I/O error occurs
     */
    public long skip(long n) throws IOException {

        for ( ; ; ) {
            // Lock on the input stream so that another thread does not
            // perform a read
            synchronized (super.in) {
                int available = 0;
                if ((available = available()) > 0) {
                    if (available > n)
                        available = (int)n;
                    return super.skip((long)available);
                }
            }
            // Wait a while before trying again
            Thread.currentThread().yield();
            try {
                Thread.currentThread().sleep(100);
            } catch (Exception e) {}
        }

    }

    /**
     * Block until the {@link #available()} method returns something other
     * than 0. This method emulates the blocking. Of course, this method
     * is not nearly as efficient as native blocking. However, it is the
     * only alternative on some Windows platforms where I/0 operations cause 
     * the entire JVM to block.
     *
     * @param b the buffer into which the data is read
     * @param off the start offset in array b at which the data is written
     * @param len the maximum number of bytes to read 
     * @return the total number of bytes read into the buffer, or -1 if the end
     *  of the stream is reached
     * @throws IOExeption if an I/O error occurs
     */
    private int readWhenAvailable(byte[] b, int off, int len)
        throws IOException
    {

        for ( ; ; ) {
            // Lock on the input stream so that another thread does not
            // perform a read 
            synchronized (super.in) {
                int available = 0;
                if ((available = available()) > 0) {
                    if (available > len)
                        available = len;
                    return super.read(b, off, available);
                }
            }
            // Wait a while before trying again
            Thread.currentThread().yield();
            try {
                Thread.currentThread().sleep(100);
            } catch (Exception e) {}
        }

    }

}
